// implements the btkcam protocol

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <sys/un.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <time.h>
#include "btkcam-client.h"


unsigned short int __htons(unsigned short int hostshort)
{
	unsigned short int netshort;
	((unsigned char *)(&netshort))[0] = (hostshort >> 8) & 0xFF;
	((unsigned char *)(&netshort))[1] = (hostshort >> 0) & 0xFF;
	return netshort;
}

// ------------------------------- btkcam client class member implementation ------------------------------

btkcam_client::btkcam_client(char *server_url, unsigned int timeout)
{	
	char cmx[50];
	status = 1;
	ready = false;
	network_timeout = timeout;
	fd=-1;

	memset(cw_0, 0, sizeof(cw_0));
	memset(cw_1, 0, sizeof(cw_1));
	
	strcpy(url, server_url);
	memset(username, 0, sizeof(username));
	memset(password, 0, sizeof(password));
	memset(hostname, 0, sizeof(hostname));
	memset(passwkey, 0, sizeof(passwkey));
	port = 0;
	if (strncmp(url,"B: ",3) != 0) return;
	parse_url();
	server_connect();
	if ( status ) return;
	ready=true;
	strcpy(cmx,"USER");
	strcat(cmx,username);
	strcat(cmx,password);
	checknet((unsigned char *)cmx,strlen(cmx));

	if ( (memcmp(answer,"ASK:NOP",7) == 0 ) || ( memcmp(answer,"ASK:",4)!=0) )
			{
			status=1;
			ready=false;
			printf("[Btkcam] Authantication Error. Server connection closed\n");
			return;
			}
	int version=0;			//client version
	int srver;			//server version
#ifdef __VERSION
	memset(cmx,0,5);
	memcpy(cmx,answer+4,3);
	srver=atoi(cmx);
	version=atoi(__VERSION);
	printf("Versions S/C  : %d - %d\n",srver,version);
#endif
	if (version!=0)	
		if (version<srver)
			{
			status=1;
			ready=false;
			printf("[Btkcam] Server Version Error. Server connection closed\n");
			return;
			}

	pcant=0;
	if (strlen(answer)>10)	
		{ 
		memcpy(provscas,answer+7,strlen(answer)-7); 
		pcant=strlen(answer)/10; 
		printf ("Provider Count: %d \n%s\n",pcant,provscas);
		}
	

	printf("[Btkcam] Sucessful Server Login\n");
	dbg = false;
}



bool btkcam_client::is_ready( void )
{
	return ready;
}

void btkcam_client::checknet(unsigned char *data, int lent)
{
	unsigned char qmsg[2048];
	int netdata,ixx;
	char prf[50];
	
	//if (dbg) printf("ORG TCP Command : >>%s%s<< \n", qmessage,params);
	if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &optval, &optlen) < 0)
  		{
		status=1;
		ready=false;
		perror("[Btkcam] Error in socket. Getsocketoption \n");
    		return;
 		}
	if (!ready) {
		status=1;
		printf("[Btkcam] Protocol is down.\n");
		return;
		}

	memcpy(qmsg,data,lent);
	netdata=lent;
	status=0;

	memset(prf,0,sizeof(prf));
	if (memcmp(qmsg,"ECM:",4)==0) memcpy(prf,qmsg,22);
			else memcpy(prf,qmsg,netdata);
	if (dbg) printf("[Btkcam] TX CMD >%s>%d\n", prf,netdata);
	
	/*check route for ECM
	if (memcmp(qmsg,"ECM:",4)==0) 
		{
		for (ixx=0;ixx<pcant;ixx++)
		if (memcmp(prf+4,provscas+(ixx*10),10)==0)  break;
		if (ixx==pcant) 
			{
			memcpy(answer,"ASK:NOP\0",8);
			if (dbg) printf("[Btkcam] Rx CmD >%s<%d\n",answer,strlen(answer));
			return;
			}
		}	*/
			
	if (encryption) netdata=btkcrypt(qmsg,netdata,passwkey);
	
	netdata = write(fd,qmsg,netdata); 
	//else crypt    	
	if (netdata < 0) 
        	{
		status = 1;
		ready=false;
		perror("[Btkcam] Error writing to socket.\n");
		return;
		}
	
	FD_ZERO(&fdset);
  	FD_SET(fd, &fdset);
	tval.tv_sec = network_timeout;
	tval.tv_usec=0;
	fdret = select(fd+1, &fdset, NULL, NULL, &tval);    //5 seconds to Read
	if (fdret < 0)
		{
		perror("[Btkcam] Connection Error.\n");	
		status=1;
		ready=false;
		return;
		}

	// Read Answer from Server	
	if FD_ISSET(fd,&fdset) 
		{
		netdata = read(fd,qmsg,2048);
		if (netdata < 0) 
        		{
			status = 1;
			ready=false;
			perror("[Btkcam] Network Read Failed.\n");
			return;
			}
		if (netdata == 0) 
        		{
			status = 1;
			ready=false;
			perror("[Btkcam] .0. data received from socket.\n");
			return;
			}
		if (encryption) netdata=btkdecrypt(qmsg,netdata,passwkey);
		memset(answer,0,sizeof(answer));
		memcpy(answer,qmsg,netdata);
		if (dbg) printf("[Btkcam] RX CMD >%s<%d\n",answer,strlen(answer));
		} 
	else
		{
		status = 1;
		ready=false;
		strcpy(answer,"ERR:timeout");
		printf("Timeout ... No answer\n");
		}
}

void btkcam_client::parse_url( void )
{		
	sscanf(url, "B: %s %d %16s %16s  %16s",  hostname, &port, username, password, passwkey);
	
		printf("[Btkcam] User : %s \n", username);
		printf("[Btkcam] Pass : %s %s\n", password, passwkey);
		printf("[Btkcam] Host : %s %d\n", hostname, port);
		encryption=false;
		int kewl;
		kewl=strlen(passwkey);
		if ( kewl==16 ) 	encryption=true;
		printf("[Btkcam] Encryption is %d\n", encryption );
}
	
btkcam_client::~btkcam_client( void )
{
	server_disconnect();
}

void btkcam_client::server_connect( void )
{
	int oval,netdata;
   	socklen_t olen;
	struct hostent *host_address;
	struct sockaddr_in socket_address;
	unsigned char buffer[300];
		
	host_address = gethostbyname(hostname);
	if(!host_address)
	{
		status = 1;
		printf("[Btkcam] Could not resolve hostname.\n");
		return;
	}


	if((fd = socket(PF_INET, SOCK_STREAM, 0)) < 0)
	{
		status = 1;
		perror("[Btkcam] Could not get socket descriptor.\n");
		return;
	}
	
	socket_address.sin_family = AF_INET;
    	socket_address.sin_port = __htons(port);
    	socket_address.sin_addr.s_addr = ((struct in_addr *) host_address->h_addr)->s_addr;

	if( connect(fd, (struct sockaddr *)(&socket_address), sizeof(struct sockaddr)) < 0 )
	{
		status = 1;
		perror("[Btkcam] Failed to connect to remote server.\n");
		return;
	}
	printf("[Btkcam] Successfully connected to server.\n");

	// Welcome message recieve
	FD_ZERO(&fdset);
  	FD_SET(fd, &fdset);

	tval.tv_sec = network_timeout;
	tval.tv_usec=0;

	fdret = select(fd+1, &fdset, NULL, NULL, &tval);    //5 seconds to connect and get initial answer
	if (fdret < 0)
		{
		perror("[Btkcam] Connection Error.\n");	
		status=1;
		return;
		}
		
	memset(buffer, 0, sizeof(buffer));
	netdata = read(fd,buffer,255);
	if (netdata <= 0) 
        	{
		status = 1;
		perror("[Btkcam] Error reading from socket.\n");
		return;
		}
	memcpy(answer,buffer,netdata);
	printf("[Btkcam] init : %s\n",answer);
	status=0;

}

void btkcam_client::server_disconnect( void )
{
	if (fd>=0) 
		{
		if (shutdown(fd, 2)<0) perror("Socket shutdown");
		if (close(fd)<0) perror("Socket close");
		}
	ready = false;
	printf("[Btkcam] Server Connection Closed : %d\n",fd);
	fd=-1;
}


//btkcrypt  CRYPT
int btkcam_client::btkcrypt (unsigned char* message, int msglen, char *keyiv )
{
unsigned char crp[2048];
int	ib,jb,zb;
unsigned char 	rand1,rand2,rand3;
int randx;
char		*src;
unsigned char 	crkeys[8];
unsigned char 	*dst = crkeys;
unsigned char	*end = crkeys + sizeof(crkeys);
unsigned int 	tmpu;
						
src = keyiv;	//Convert hexchar array to cw0[8], cw1[8] bytes
dst = crkeys;
end = crkeys + sizeof(crkeys);
while (dst < end && sscanf(src, "%2x", &tmpu) == 1)
	{
	*dst++ = tmpu;
	src += 2;
	}

memcpy(crp,message,msglen);

srand ( time(NULL) );
rand1 = rand() % 256;
rand2 = rand() % 256;
rand3 = rand2 & 0x7;

crp[msglen]=rand1;		// add rand1 to the end of crp
for (ib=0;ib<msglen;ib++) crp[ib]=crp[ib]^rand1;  //xor crp except rand1
for (ib=0;ib<msglen+1;ib++) 			 //rotate all using rand3
	crp[ib]=( ( crp[ib] >> rand3 ) | ( crp[ib] << (8-rand3) ) );
for (ib=0;ib<msglen+1;ib++) crp[ib]=crp[ib]^rand2;  //xor with rand2
crp[msglen+1]=rand2;
for (ib=0;ib<msglen+2;ib++)
	crp[ib]=crp[ib] ^ crkeys[ib % 8];
memcpy(message, crp,msglen+2);
return msglen +2;
}


//btk decrypt functionc
int btkcam_client::btkdecrypt (unsigned char* message, int msglen, char *keyiv )
{
unsigned char crp[2048];
int	ib,jb,zb;
unsigned char 	rand1,rand2,rand3;
int randx;
char		*src;
unsigned char 	crkeys[8];
unsigned char 	*dst = crkeys;
unsigned char	*end = crkeys + sizeof(crkeys);
unsigned int 	tmpu;
						
src = keyiv;	//Convert hexchar array to cw0[8], cw1[8] bytes
dst = crkeys;
end = crkeys + sizeof(crkeys);
while (dst < end && sscanf(src, "%2x", &tmpu) == 1)
	{
	*dst++ = tmpu;
	src += 2;
	}

memcpy(crp,message,msglen);
for (ib=0;ib<msglen;ib++)			//xor with keys
	crp[ib]=crp[ib] ^ crkeys[ib % 8];

rand2=crp[msglen-1];			// last char is rand
rand3=rand2 & 0x7;
for (ib=0;ib<msglen;ib++)		//xor with rand2
	crp[ib]=crp[ib]^rand2;
for (ib=0;ib<msglen;ib++) 		//rotate all using rand3
	crp[ib]=( ( crp[ib] << rand3 ) | ( crp[ib] >> (8-rand3) ) );
rand1=crp[msglen-2];
for (ib=0;ib<msglen-1;ib++)		//xor with rand2
	crp[ib]=crp[ib]^rand1;
crp[msglen-1]=0;
memcpy(message,crp,msglen-1);
return msglen - 2;
}

